import logging
import subprocess
import shutil
import time
import tempfile
from pathlib import Path

import PIL.Image

import metaindex.configuration
import metaindex.shared
from metaindex import Cache
from metaindex import proto
from metaindex.client import Client

from filecabinet import utils
from filecabinet.cabinet import Cabinet, Document


METAINDEXCONFIG = '''[General]
cache = {cache}
socket = {socket}
index-unknown = no
preferred-sidecar-format = .yaml
recursive-extra-metadata = no
implicit-tags = issuer, creator, title, tag, subject
ocr = no

[Indexer:rule-based]
{rules}

[Synonyms]
fulltext = pdf.fulltext, extra.fulltext
issuer = extra.issuer
creator = extra.creator
subject = extra.subject
tag = extra.tag, extra.tags, rules.tag, rules.tags
language = extra.language, rules.language
date = extra.date, rules.date
checksum = sha256

[Server]
loglevel = warning
logfile = {logfile}
index-new-tags = yes
'''


class FilecabinetCache(Cache):
    """A cache with different 'reduce' rules"""
    def __init__(self, manager, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.manager = manager

    def reduce(self, entry):
        for key in set(entry.metadata.keys()):
            values = entry.metadata[key]
            skey = None

            if key == 'sha256':
                skey = key
            elif key in self.config.synonymized and any(len(str(v.raw_value)) > 0 for v in values):
                skey = 'extra.' + self.config.synonymized[key]
            elif key.startswith('extra.'):
                skey = key.lower()
            elif key.startswith('rules.') and len(key.strip()) > 6:
                skey = 'extra.' + key.strip()[6:]

            if skey is not None:
                if skey not in entry.metadata:
                    entry.metadata[skey] = []
                entry.metadata[skey] = list(set(entry.metadata[skey] + [v for v in values if len(str(v.raw_value)) > 0]))

            if key != skey:
                del entry.metadata[key]

        return entry

    def find(self, *args, **kwargs):
        results = [Document.wrap(self.manager.path_cabinet.get(Path(*entry.path.parts[:-4])), entry)
                   for entry in super().find(*args, **kwargs)]
        return results

    def get(self, *args, **kwargs):
        results = [Document.wrap(self.manager.path_cabinet.get(Path(*entry.path.parts[:-4])), entry)
                   for entry in super().get(*args, **kwargs)]
        return results

    def shutdown(self):
        self.client.send(proto.Shutdown())


class Manager:
    def __init__(self, configuration):
        self.config = configuration

        self.ocr = None

        # check whether there is any OCR, if needed
        if self.config.bool('OCR', 'enabled'):
            self.ocr = shutil.which('tesseract')

        # create the metaindex configuration, overwrite existing ones
        cachepath = self.config.path('General', 'database')
        rules = []
        if 'Rules' in self.config:
            rules = [f"{k} = {self.config.get('Rules', k)}" for k in self.config['Rules']]
        self.miconfig = cachepath / 'metaindex.conf'
        if not self.miconfig.is_file() or (self.config.filepath is not None
           and self.config.filepath.is_file() and
           metaindex.shared.get_last_modified(self.config.filepath) > metaindex.shared.get_last_modified(self.miconfig)):
            cachepath.mkdir(parents=True, exist_ok=True)
            self.miconfig.parent.mkdir(parents=True, exist_ok=True)
            socketpath = self.miconfig.parent / 'metaindex.sock'
            logfile = self.miconfig.parent / 'metaindex.log'
            self.miconfig.write_text(METAINDEXCONFIG.format(cache=str(cachepath),
                                                            socket=str(socketpath),
                                                            logfile=str(logfile),
                                                            rules="\n".join(rules)))
        self.client_config = metaindex.configuration.load(self.miconfig)
        self.cache = None  # lazily created when a new session is created

        self.cabinets = []
        self.path_cabinet = {}  # mapping cabinet.basedir -> cabinet
        for cabinetname in configuration.cabinets():
            try:
                cabinet = Cabinet(cabinetname, self)
                self.cabinets.append(cabinet)
                self.path_cabinet[cabinet.basedir] = cabinet
            except RuntimeError as exc:
                logging.error(exc)
                continue

    def launch_server(self):
        """Launch the corresponding metaindexserver"""
        client = Client(self.client_config)
        if not client.connection_ok():
            subprocess.run(['metaindexserver', '-c', str(self.miconfig), '-d'])
            time.sleep(1)
        assert client.connection_ok()

    def find_cabinet(self, name):
        for cabinet in self.cabinets:
            if cabinet.name == name:
                return cabinet
        return None

    def new_session(self):
        if self.cache is None:
            self.cache = FilecabinetCache(self, self.client_config)
        return Session(self)

    def process_image(self, fn, default_lang=None):
        """Run Tesseract OCR on the given file and create a OCR'ed PDF

        Returns the ``Path`` to the created PDF file or ``None`` upon error.
        """
        if self.ocr is None:
            return None

        dpi = PIL.Image.open(fn).info.get('dpi', [0])[0]

        with tempfile.NamedTemporaryFile(suffix='.pdf', dir=fn.parent, delete=True) as tmpfd:
            pdffile = Path(tmpfd.name)
            outputbase = str(pdffile.parent / pdffile.stem)

        lang = utils.get_language_from_filename(fn)
        if lang is None:
            lang = default_lang or self.config.get('OCR', 'default-language', 'eng')

        result = subprocess.run([self.ocr,
                                 str(fn),
                                 outputbase,
                                 '-c', 'tessedit_create_pdf=1',
                                 '-c', f'user_defined_dpi={dpi}',
                                 '-l', lang],
                                stdout=subprocess.DEVNULL,
                                stdin=subprocess.DEVNULL,
                                stderr=subprocess.DEVNULL,
                                check=False)
        if result.returncode == 0:
            return pdffile
        return None


class Session:
    def __init__(self, manager):
        self.manager = manager
        self.cache = manager.cache

    def pickup(self, cabinet=None):
        if cabinet is None:
            for that in self.manager.cabinets:
                self.pickup(that)
        else:
            cabinet.pickup(self)

    def reindex(self, cabinet=None):
        if cabinet is None:
            for that in self.manager.cabinets:
                self.reindex(that)
        else:
            cabinet.reindex(self)

    def find(self, query):
        """Return a list of all documents that match this query"""
        return list(self.cache.find(query))

    def update(self, doc):
        """Update the stored metadata of 'doc'"""
        cabinet = self.manager.path_cabinet.get(doc.path.parent.parent.parent.parent)
        if cabinet is None:
            logging.error(f"Document {doc.path} does not appear to be part of any cabinet.")
            return

        cabinet.update(self, [doc])

    def add(self, cabinet, sources, samedoc=False, allowduplicates=False, metadata={}):
        """Add the source files to the cabinet

        :arg cabinet The target cabinet to add the files to
        :arg sources A list of files
        :arg samedoc If True, add all files to the same document as pages, otherwise create new documents per file
        :arg allowduplicates If False, reject any duplicate documents/pages. If True, accept.
        :arg metadata Metadata to apply to all added documents
        """



import configparser
import os
import logging
import pathlib

from . import utils


HERE = pathlib.Path(__file__).parent

HOME = pathlib.Path().home()
PROGRAMNAME = 'filecabinet'
CONFFILENAME = PROGRAMNAME + os.path.extsep + 'conf'
CONFIGFILE = HOME / ".config" / PROGRAMNAME / CONFFILENAME
INDEXPATH = HOME / ".cache" / PROGRAMNAME


try:
    from xdg import BaseDirectory
    CONFIGFILE = pathlib.Path(BaseDirectory.load_first_config(CONFFILENAME) or CONFIGFILE)
    INDEXPATH = pathlib.Path(BaseDirectory.save_cache_path(PROGRAMNAME) or INDEXPATH)
except ImportError:
    BaseDirectory = None


CONF_DEFAULTS = {'General': {'loglevel': 'warning',
                             'database': INDEXPATH,
                            },
                 'Shell': {'document_opener': 'xdg-open',
                          },
                 'OCR': {'enabled': 'yes',
                         'default-language': 'eng',
                        },
                 }


class Configuration:
    def __init__(self, conf):
        self.conf = conf
        self.filepath = None
        """Path to the configuration file"""
        self._userfile = None

    def cabinets(self):
        for group in self.conf:
            if group not in CONF_DEFAULTS.keys() and self.conf[group].get('path', None) is not None:
                yield group

    def __contains__(self, group):
        return group in self.conf

    def __getitem__(self, group):
        return self.conf[group]

    def get(self, group, item, default=None):
        return self.conf[group].get(item, default)

    def bool(self, group, item, default='n'):
        return self.get(group, item, default).lower() in ['y', 'yes', '1', 'true', 'on']

    def number(self, group, item, default='0'):
        value = self.get(group, item, default)
        if value.isnumeric():
            return int(value)
        return None

    def size(self, group, item, default='0'):
        return utils.parse_size(self.get(group, item, default))

    def duration(self, group, item, default=''):
        return utils.parse_duration(self.get(group, item, default))

    def path(self, group, item, default=None):
        return pathlib.Path(self.get(group, item, default)).expanduser()

    def list(self, group, item, default='', separator=',', strip=True, skipempty=True):
        result = []
        for v in self.get(group, item, default).split(separator):
            if strip:
                v = v.strip()
            if skipempty and len(v) == 0:
                continue
            result.append(v)
        return result


def get_configuration(conffile=None):
    conf = configparser.ConfigParser(interpolation=None)
    conf.read_dict(CONF_DEFAULTS)

    if conffile is not None and not os.path.exists(os.path.abspath(conffile)):
        logging.warning(f"Configuration file {conffile} not found. "
                        "Using defaults.")
        conffile = None
    if conffile is None and CONFIGFILE.is_file():
        conffile = CONFIGFILE

    if conffile is not None:
        logging.info(f"Loading configuration from {conffile}.")
        conf.read([conffile])

    finalconfig = Configuration(conf)
    finalconfig.filepath = pathlib.Path(conffile or CONFIGFILE)
    return finalconfig



import unittest
import re

import filecabinet.processor as fp


class TestRuleParser(unittest.TestCase):
    def test_one_rule(self):
        text = "match /hello/"
        rules = fp.Rule.parse(text)
        self.assertEqual(len(rules), 1)
        self.assertEqual(len(rules[0].conditions), 1)
        self.assertTrue(rules[0].match("why hello there"))

    def test_no_match(self):
        text = "match /sausage/"
        rules = fp.Rule.parse(text)
        self.assertFalse(rules[0].match("why hello there"))

    def test_fail_parse_match(self):
        text = "match /spam"
        with self.assertRaises(fp.ParseError):
            fp.Rule.parse(text)

    def test_more_conditions(self):
        text = "match /albatros/ and /foo/ /ice cream/"
        rules = fp.Rule.parse(text)
        self.assertTrue(rules[0].match("You fool, I don't want an albatros, I want ice cream!"))

    def test_impossible_expression(self):
        text = "match /some\/ stuff/"
        with self.assertRaises(fp.ParseError):
            fp.Rule.parse(text)

    def test_possible_expression(self):
        text = "match :some/ stuff:"
        fp.Rule.parse(text)

    def test_many_rules(self):
        text = "match /banana/\nmatch /spamalot/"
        rules = fp.Rule.parse(text)
        self.assertEqual(len(rules), 2)

    def test_find_rules(self):
        text = """match /meh/
find issuer /(Big Corp [a-zA-Z]+)/
set issuer {issuer}

match /fibsh/
  find format /Format[ \\n\\t]+([^\\n]+)/
  find title /Subject[ \\n\\t]+([^\\n]+)/
  set format {format}
  set title {title}
"""
        rules = fp.Rule.parse(text)
        self.assertEqual(len(rules), 2)
        self.assertEqual(len(rules[0].finds), 1)
        self.assertEqual(len(rules[1].finds), 2)

    def test_regex_fail(self):
        text = "match /bla [/"

        with self.assertRaises(fp.RegExParseError):
            fp.Rule.parse(text)

    def test_unused_finds(self):
        text = """match /meh/
find foo /(bar)/
"""

        with self.assertRaises(fp.UnusedFindsError):
            fp.Rule.parse(text)

    def test_undefined_find(self):
        text = """match /meh/
set state very {state}
"""

        with self.assertRaises(fp.UndefinedFindError):
            fp.Rule.parse(text)

    def test_set_regex(self):
        text = """match /foo/ and /bar/
set issuer /(Big Corp [a-zA-Z]+)/"""

        rules = fp.Rule.parse(text)
        self.assertTrue(isinstance(rules[0].sets['issuer'].pop(), re.Pattern))

    def test_spaced_set(self):
        text = """match /albatross/
set "ice cream" yes
set "ice cream flavour" /([a-z]+) taste/"""

        rules = fp.Rule.parse(text)
        self.assertEqual(rules[0].sets['ice cream'].pop(), 'yes')


class TestRuleExecution(unittest.TestCase):
    GENERIC_INVOICE = """Invoice #12345\n2012-04-13\n\nDear Customer,\nThanks for ordering at Big Corp.
Please pay us now this ridiculous amount: $500.20 until 2012-05-30.\nThanks."""

    def test_str_set(self):
        rules = fp.Rule.parse("""match /Big Corp/
set issuer Big Corp Inc.
""")

        attrs = rules.run(TestRuleExecution.GENERIC_INVOICE)
        self.assertEqual(attrs['issuer'], {'Big Corp Inc.',})

    def test_find_set(self):
        rules = fp.Rule.parse("""match /Invoice/
find invoicenr /Invoice[ \t]+#([0-9]+)/
find amount /[ \t]+\$[ \t]*([0-9]+\.?[0-9]{2}?)/
set bill "{invoicenr}"
set amount ${amount}
# set other "not"
; set commented out
""")

        attrs = rules.run(TestRuleExecution.GENERIC_INVOICE)
        self.assertEqual(len(attrs), 2)
        self.assertEqual(attrs['bill'], {'12345',})
        self.assertEqual(attrs['amount'], {'$500.20',})

    def test_direct_regex(self):
        rules = fp.Rule.parse("""match /Big Corp/
set invoicenr /Invoice[ \t]+#([0-9]+)/
set date /([0-9]{4}-[01][0-9]-[0-3][0-9])/
""")

        attrs = rules.run(TestRuleExecution.GENERIC_INVOICE)
        self.assertEqual(attrs['invoicenr'], {'12345',})
        self.assertEqual(attrs['date'], {'2012-05-30', '2012-04-13'})

    def test_multi_set(self):
        rules = fp.Rule.parse("""match /Big Corp/
set issuer Big Corp
set issuer BCinc""")

        attrs = rules.run(TestRuleExecution.GENERIC_INVOICE)
        self.assertEqual(attrs['issuer'], {'Big Corp', 'BCinc'})

    def test_case_insensitive_match(self):
        rules = fp.Rule.parse("""match .big corp.i
set issuer Big Corp
""")
        attrs = rules.run(TestRuleExecution.GENERIC_INVOICE)
        self.assertEqual(attrs['issuer'], {'Big Corp'})

    def test_case_insensitive_find(self):
        rules = fp.Rule.parse("""match /Big Corp/
find issuer /(big [a-z]+)/i
set issuer {issuer}
""")
        attrs = rules.run(TestRuleExecution.GENERIC_INVOICE)
        self.assertEqual(attrs['issuer'], {'Big Corp'})

    def test_case_insensitive_set(self):
        rules = fp.Rule.parse("""match /Big Corp/
set issuer /(big [a-z]+)/i
""")
        attrs = rules.run(TestRuleExecution.GENERIC_INVOICE)
        self.assertEqual(attrs['issuer'], {'Big Corp'})

    def test_spaced_set(self):
        text = """match /Big Corp/
set "silly amount" /\\$([0-9]+)/"""

        rules = fp.Rule.parse(text)
        attrs = rules.run(TestRuleExecution.GENERIC_INVOICE)
        self.assertEqual(attrs['silly amount'], {'500'})


if __name__ == "__main__":
    unittest.main()




"""An interactive commandline shell to work with filecabinet"""
import subprocess
import shutil
import shlex
import os
import tempfile

try:
    import readline
except ImportError:
    readline = None

import yaml


class Shell:
    def __init__(self, session):
        self.session = session
        self.manager = session.manager

        self.quit = False
        self.listed_docs = []

        self.opener = shutil.which(self.manager.config.get('Shell', 'document_opener', 'xdg-open'))
        if self.opener is not None:
            self.opener = shlex.split(self.opener)

        editor = self.manager.config.get('Shell', 'editor', None)
        if editor is None:
            for name in ['VISUAL', 'EDITOR']:
                value = os.getenv(name)
                if value is None:
                    continue
                editor = shutil.which(value)
                if editor is not None:
                    break
        self.editor = editor
        if self.editor is not None:
            self.editor = shlex.split(self.editor)

        self.commands = {'quit': self.do_quit,
                         'help': self.help,
                         'info': self.do_info,
                         'open': self.do_open,
                         'edit': self.do_edit,
                         'pickup': self.do_pickup,
                         'seen': self.do_seen,
                         'new': self.do_new,
                         'kill': self.do_stop,
                         'compact': self.do_compact,
                         'find': self.do_find,
                         'list': self.do_find,
                         }

    def do_quit(self, _):
        """Exit the shell"""
        self.quit = True

    def help(self, _):
        """Show this help"""
        for cmd in sorted(self.commands.keys()):
            fnc = self.commands[cmd]
            print(f"{cmd}\t{fnc.__doc__}")

    def do_find(self, args):
        """Find documents or list the most recently found documents"""
        if len(args) > 0:
            query = " ".join(args)
            self.listed_docs = list(self.session.find(query))
        self.print_doc_list()

    def print_doc_list(self):
        idxwidth = len(str(len(self.listed_docs)))
        for idx, doc in enumerate(self.listed_docs):
            number = f"({idx+1: >{idxwidth}})"
            is_new = "*" if doc.is_new() else " "
            title = doc.path.stem
            for tag, value in doc:
                if not tag.endswith('.title'):
                    continue
                title = value
                break

            tags = ' '.join([f'#{value}'
                             for tag, value in doc
                             if tag.endswith('.tag') and value != 'new'])

            print(f"{number} {is_new} {title} {tags}")

    def do_open(self, args):
        """Open the document in an external program"""
        if self.opener is None:
            print("No opener defined")
            return

        doc = self.doc_from_list(args)
        if doc is None:
            print("Usage: open document-number")
            return

        subprocess.Popen(self.opener + [str(doc.path)],
                         shell=False,
                         stdout=subprocess.DEVNULL,
                         stdin=subprocess.DEVNULL,
                         stderr=subprocess.DEVNULL)

    def do_edit(self, args):
        """Edit the metadata in an external editor"""
        if not isinstance(self.editor, list):
            print("No editor defined.")
            return

        doc = self.doc_from_list(args)
        if doc is None:
            print("Usage: edit document-number")
            return

        sidecar = doc.path.parent / (doc.path.stem + '.yaml')

        with tempfile.NamedTemporaryFile('w+t', encoding='utf-8', suffix='.yaml') as tmpfh:
            tmpfh.write(sidecar.read_text(encoding='utf-8'))
            tmpfh.flush()

            subprocess.run(self.editor + [tmpfh.name], check=False)
            tmpfh.flush()
            tmpfh.seek(0)

            try:
                other = yaml.safe_load(tmpfh.read())
            except yaml.YAMLError as exc:
                print(f"Failed to load metadata file: {exc}")
                return

            doc.metadata = {}
            for key, values in other.items():
                if not isinstance(values, list):
                    values = [values]
                for value in values:
                    doc.add('extra.' + key, value)

            self.session.update(doc)

    def do_info(self, args):
        """Show the metadata of the selected document"""
        doc = self.doc_from_list(args)
        if doc is None:
            print("Usage: info document-number")
            return

        skip = {'extra.fulltext',}
        tagwidth = max(len(tag) for tag, _ in doc if tag not in skip)

        for tag, value in doc:
            if tag in skip:
                continue
            print(f"{tag: <{tagwidth}}    {value}")

    def doc_from_list(self, args):
        if len(args) < 1:
            return None
        if not args[0].isnumeric():
            return None
        nr = int(args[0])-1
        if nr >= len(self.listed_docs) or nr < 0:
            return None
        return self.listed_docs[nr]

    def do_pickup(self, _):
        """Process all files in all cabinets inboxes"""
        self.session.pickup()
        self.do_find(['tag:new'])

    def do_stop(self, _):
        """Quit the shell and also stop the background server"""
        self.quit = True
        self.manager.cache.shutdown()

    def do_compact(self, _):
        """Compact the backend database"""
        self.manager.cache.compact()

    def do_seen(self, args):
        """Mark the document as seen (remove 'new' tag)"""
        doc = self.doc_from_list(args)
        if doc is None:
            print("Usage: seen document-number")
            return

        if not doc.is_new():
            return

        doc.delete(('extra.tag', 'new'))
        self.session.update(doc)

    def do_new(self, args):
        """Mark that document as new"""
        doc = self.doc_from_list(args)
        if doc is None:
            print("Usage: new document-number")
            return

        if doc.is_new():
            return

        doc.add('extra.tag', 'new')
        self.session.update(doc)

    def run(self):
        if self.opener is None:
            print("[Warning] There is no document opener defined or the "
                  "defined document opener could not be found.\nThe 'open' "
                  "command will not work.\nPlease "
                  "check your configuration file to fix this.")
        if self.editor is None:
            print("[Warning] There is no metadata editor defined or the "
                  "defined editor could not be found.\nThe 'edit' "
                  "command will not work.\nPlease "
                  "check your configuration file to fix this.")
        while not self.quit:
            try:
                rawcmd = input('> ')
            except (KeyboardInterrupt, EOFError):
                self.quit = True
                continue

            cmds = tokenize(rawcmd)
            matches = []
            if len(cmds) > 0:
                for name in sorted(self.commands.keys()):
                    if name.startswith(cmds[0]):
                        matches.append(name)
            if len(matches) == 1:
                self.commands[matches[0]](cmds[1:])
            elif len(matches) > 1:
                print("Not quite sure what of these commands you meant:\n {}".format("\n ".join(matches)))
            elif len(rawcmd.strip()) > 0:
                print("I don't know what you mean. Try 'help' to see all available commands.")


def tokenize(text):
    words = []

    word = ''
    quote = None
    escaped = False

    for letter in text:
        if escaped:
            word += letter
            escaped = False
        else:
            if letter == '\\':
                escaped = True
            elif quote is not None and letter == quote:
                quote = None
                words.append(word)
                word = ''
            elif quote is None and letter in " \t":
                words.append(word.strip())
                word = ''
            elif quote is None and letter in '"\'' and len(word.strip()) == 0:
                word = ''
                quote = letter
            else:
                word += letter

    if len(word.strip()) > 0:
        words.append(word.strip())

    return [word for word in words if len(word) > 0]
